//------------------------------------------------------------------------------
// Силовой модуль. Контроллер клапана отбора
// Список изменений - в конце файла
// OldBean, 05.02.2018
//------------------------------------------------------------------------------
#include <avr/io.h>
#include <util/twi.h>
#include <avr/interrupt.h>

#define F_CPU 16000000UL
#define ADDR 0x14      // Адрес устройства на шине I2C
                         
volatile uint16_t pwm = 0;  // Заданный ШИМ в полупериодах (десятых долях %)
volatile uint16_t wpwm;     // Рабочая переменная
static volatile uint8_t index = 0; // Номер байта (0 - младший, 1 - старший)
volatile uint16_t cntr = 0; // Счетчик полупериодов
//------------------------------------------------------------------------------
ISR(TWI_vect) { // Обработчик событий модуля TWI
  uint16_t w;
  switch(TW_STATUS) {    // Статус - в 5 старших битах регистра TWSR
    case TW_SR_DATA_ACK: // Данные приняты, ACK отправлен - 0x80
      w = TWDR;
      if(index) { // Принят старший байт. Значит формируем новое значение pwm
        w <<= 8; pwm = wpwm | w; index = 0;
        if(pwm > 1000) pwm = 1000;
      } else { // Принят младший байт - формируем младшие биты будущего pwm
        wpwm = w; index = 1;
      }
      break;
    case TW_ST_SLA_ACK:  // Запрос SLA+R, ACK отправлен - 0xA8
      TWDR = 0xff;       // По запросу - отправляем малинке кучу единичек (OK),
      index = 0;         // что означает, что контроллер готов принять два
      break;             // последовательных байта (начиная с младшего)
    case TW_BUS_ERROR:   // Неверные условия "старт" или "стоп" - 0x00
      TWCR = _BV(TWIE) | _BV(TWINT) | _BV(TWEA) | _BV(TWEN); // "Сброс" TWI
      break;
  }
  TWCR |= _BV(TWINT);   // Флаг TWINT здесь НЕ очищается автоматически! Очистим.
} 
//------------------------------------------------------------------------------
ISR(INT0_vect) { // Внешнее прерывание INT0 (Zero от детектора нуля)
  cntr++; // Считаем полупериоды
  if(cntr == 1000) { // 10 секунд (1000 полупериодов) прошло
    cntr = 0;   // Сбрасываем счетчик
    if(pwm > 0) // Включаем клапан (начало импульса ШИМ)
      PORTD |= _BV(PD0);
  }
  if(cntr >= pwm && (PORTD & _BV(PD0))) // Пора выключать клапан
    PORTD &= ~_BV(PD0);
  if(cntr < pwm && (~PORTD & _BV(PD0))) // Если pdm был изменен, но время
    PORTD |= _BV(PD0);                  // еще не истекло - откроем клапан
}
//------------------------------------------------------------------------------
int main(void) {
  DDRD = _BV(PD0); // Пин PD0 - на вывод (управление оптосимистором)
  // Инициализация TWI-модуля --------------------------------------------------
  TWAR = ADDR << 1;   // Адрес устройства - в старших 7 битах
  TWCR = _BV(TWIE);   // Разрешаем прерывания от TWI
  TWCR |= _BV(TWEA);  // Разрешаем подачу импульса ACK
  TWCR |= _BV(TWINT); // Очистим флаг TWINT (очищается при установке в 1)
  TWCR |= _BV(TWEN);  // Разрешаем работу TWI и активирует TWI-интерфейс
  // Инициализация внешнего прерывания INT0
  EICRA = _BV(ISC01) | _BV(ISC00); // Прерывания по фронту на пине PD2/INT0
  EIMSK = _BV(INT0); // Разрешаем внешние прерывания на пине PD2/INT0
  
  sei(); // Разрешаем глобальные прерывания
  while(1) { // Бесконечный цикл
  }
  return 0;
}
//------------------------------------------------------------------------------
// Список изменений
// 1. 05.02.2018 - Отлажен базовый вариант.
//------------------------------------------------------------------------------

